; ADC  
; This is part of the Z80-CP/M emulator written by Sprite_tm.
;
;    Copyright (C) 2013 Leo C.
;
;    This file is part of avrcpm.
;
;    avrcpm is free software: you can redistribute it and/or modify it
;    under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    avrcpm is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with avrcpm.  If not, see <http://www.gnu.org/licenses/>.
;
;    $Id: $
;

#if ADC_SUPPORT

;
;ADMUX:
;	REFS1:0		1 = AVCC, 3 = 1.1V (VBG)
;	ADLAR		1 = ADC left justified
;	MUX3:0		6 = Ch6, 7 = Ch7, (8=Temp, 14=BG)
;
;ADCSRA:
;	ADEN		ADC Enable
;	ADSC		Start Conversation
;	ADIF		ADC Interrupt Flag
;	ADIE		ADC Interrupt Enable
;	ADPS2:0		7 = Division Factor 128
;
;ADCL	ADC Data Register
;ADCH
;

;--------------------------------------------------------
; Read 8 Bit from ADC, Channel 6 or Channel 7
;
; temp3: 0 = Channel 6
;	 1 = Channel 7
;	 2 = Channel 8	= VTemp
;

adc_read8:
	ldi	temp2,6
	rjmp	adc_r81
adc_r80:
	dec	temp3
	inc	temp2
adc_r81:
	cpse	temp3,_0
	rjmp	adc_r80

	sbrc	temp2,3				;Temp sensor?
	ori	temp2,(1<<REFS1)
	ori	temp2,(1<<REFS0) | (1<<ADLAR)
	outm8	ADMUX,temp2
	ldi	temp2,(1<<ADEN)|(1<<ADSC)|(7<<ADPS0)
	outm8	ADCSRA,temp2
adc_r8wait:
	inm8	temp,ADCSRA
	sbrs	temp,ADIF
	rjmp	adc_r8wait

	inm8	temp,ADCH
	ldi	temp2,(1<<ADEN)|(1<<ADIF)|(7<<ADPS0)	;clear pending int.
	outm8	ADCSRA,temp2

	ret

;--------------------------------------------------------
; Measure 1.1V bandgap voltage (VBG), with VCC as reference.
; This can be used to measure VCC, if VBG is known.
;
; VCC = VBG * 1024 / ADC 
;
; temp3: 0 = Get Value of ADCL
;	 1 = Get Value of ADCH
;

	.dseg
adcvcch:
	.byte	1

	.cseg
adc_readvcc:
	cpse	temp3,_0
	rjmp	adc_rvcc_high

	ldi	temp2,(1<<REFS0) | 14 ;
	outm8	ADMUX,temp2
	ldi	temp2,(1<<ADEN)|(1<<ADSC)|(7<<ADPS0)
	outm8	ADCSRA,temp2
adc_rvcc_wait:
	inm8	temp,ADCSRA
	sbrc	temp,ADSC
	rjmp	adc_rvcc_wait

	inm8	temp,ADCL
	inm8	temp2,ADCH
	sts	adcvcch,temp2
	ldi	temp2,(1<<ADEN)|(1<<ADIF)|(7<<ADPS0)	;clear pending int.
	outm8	ADCSRA,temp2
	rjmp	adc_rvcc_ex

adc_rvcc_high:
	lds	temp,adcvcch
adc_rvcc_ex:
	ret


#endif /* ADC_SUPPORT */

